@@ -10,8 +10,6 @@ import (
 	"sync"
 	"syscall"
 	"time"
-
-	"golang.org/x/net/netutil"
 )
 
 // Server wraps an http.Server with graceful connection handling.
@@ -34,6 +32,11 @@ type Server struct {
 	// Limit the number of outstanding requests
 	ListenLimit int
 
+	// TCPKeepAlive sets the TCP keep-alive timeouts on accepted
+	// connections. It prunes dead TCP connections ( e.g. closing
+	// laptop mid-download)
+	TCPKeepAlive time.Duration
+
 	// ConnState specifies an optional callback function that is
 	// called when a client connection changes state. This is a proxy
 	// to the underlying http.Server's ConnState, and the original
@@ -41,8 +44,8 @@ type Server struct {
 	ConnState func(net.Conn, http.ConnState)
 
 	// BeforeShutdown is an optional callback function that is called
-	// before the listener is closed.
-	BeforeShutdown func()
+	// before the listener is closed. Returns true if shutdown is allowed
+	BeforeShutdown func() bool
 
 	// ShutdownInitiated is an optional callback function that is called
 	// when shutdown is initiated. It can be used to notify the client
@@ -57,6 +60,10 @@ type Server struct {
 	// Logger used to notify of errors on startup and on stop.
 	Logger *log.Logger
 
+	// LogFunc can be assigned with a logging function of your choice, allowing
+	// you to use whatever logging approach you would like
+	LogFunc func(format string, args ...interface{})
+
 	// Interrupted is true if the server is handling a SIGINT or SIGTERM
 	// signal and is thus shutting down.
 	Interrupted bool
@@ -77,6 +84,9 @@ type Server struct {
 
 	// connections holds all connections managed by graceful
 	connections map[net.Conn]struct{}
+
+	// idleConnections holds all idle connections managed by graceful
+	idleConnections map[net.Conn]struct{}
 }
 
 // Run serves the http.Handler with graceful shutdown enabled.
@@ -85,14 +95,16 @@ type Server struct {
 // If timeout is 0, the server never times out. It waits for all active requests to finish.
 func Run(addr string, timeout time.Duration, n http.Handler) {
 	srv := &Server{
-		Timeout: timeout,
-		Server:  &http.Server{Addr: addr, Handler: n},
-		Logger:  DefaultLogger(),
+		Timeout:      timeout,
+		TCPKeepAlive: 3 * time.Minute,
+		Server:       &http.Server{Addr: addr, Handler: n},
+		// Logger:       DefaultLogger(),
 	}
 
 	if err := srv.ListenAndServe(); err != nil {
 		if opErr, ok := err.(*net.OpError); !ok || (ok && opErr.Op != "accept") {
-			srv.Logger.Fatal(err)
+			srv.logf("%s", err)
+			os.Exit(1)
 		}
 	}
 
@@ -104,9 +116,10 @@ func Run(addr string, timeout time.Duration, n http.Handler) {
 // return it instead.
 func RunWithErr(addr string, timeout time.Duration, n http.Handler) error {
 	srv := &Server{
-		Timeout: timeout,
-		Server:  &http.Server{Addr: addr, Handler: n},
-		Logger:  DefaultLogger(),
+		Timeout:      timeout,
+		TCPKeepAlive: 3 * time.Minute,
+		Server:       &http.Server{Addr: addr, Handler: n},
+		Logger:       DefaultLogger(),
 	}
 
 	return srv.ListenAndServe()
@@ -128,12 +141,12 @@ func (srv *Server) ListenAndServe() error {
 	if addr == "" {
 		addr = ":http"
 	}
-	l, err := net.Listen("tcp", addr)
+	conn, err := srv.newTCPListener(addr)
 	if err != nil {
 		return err
 	}
 
-	return srv.Serve(l)
+	return srv.Serve(conn)
 }
 
 // ListenAndServeTLS is equivalent to http.Server.ListenAndServeTLS with graceful shutdown enabled.
@@ -159,9 +172,6 @@ func (srv *Server) ListenTLS(certFile, keyFile string) (net.Listener, error) {
 	if srv.TLSConfig != nil {
 		*config = *srv.TLSConfig
 	}
-	if config.NextProtos == nil {
-		config.NextProtos = []string{"http/1.1"}
-	}
 
 	var err error
 	config.Certificates = make([]tls.Certificate, 1)
@@ -170,15 +180,42 @@ func (srv *Server) ListenTLS(certFile, keyFile string) (net.Listener, error) {
 		return nil, err
 	}
 
-	conn, err := net.Listen("tcp", addr)
+	// Enable http2
+	enableHTTP2ForTLSConfig(config)
+
+	conn, err := srv.newTCPListener(addr)
 	if err != nil {
 		return nil, err
 	}
 
+	srv.TLSConfig = config
+
 	tlsListener := tls.NewListener(conn, config)
 	return tlsListener, nil
 }
 
+// Enable HTTP2ForTLSConfig explicitly enables http/2 for a TLS Config. This is due to changes in Go 1.7 where
+// http servers are no longer automatically configured to enable http/2 if the server's TLSConfig is set.
+// See https://github.com/golang/go/issues/15908
+func enableHTTP2ForTLSConfig(t *tls.Config) {
+
+	if TLSConfigHasHTTP2Enabled(t) {
+		return
+	}
+
+	t.NextProtos = append(t.NextProtos, "h2")
+}
+
+// TLSConfigHasHTTP2Enabled checks to see if a given TLS Config has http2 enabled.
+func TLSConfigHasHTTP2Enabled(t *tls.Config) bool {
+	for _, value := range t.NextProtos {
+		if value == "h2" {
+			return true
+		}
+	}
+	return false
+}
+
 // ListenAndServeTLS is equivalent to http.Server.ListenAndServeTLS with graceful shutdown enabled.
 func (srv *Server) ListenAndServeTLS(certFile, keyFile string) error {
 	l, err := srv.ListenTLS(certFile, keyFile)
@@ -197,11 +234,13 @@ func (srv *Server) ListenAndServeTLSConfig(config *tls.Config) error {
 		addr = ":https"
 	}
 
-	conn, err := net.Listen("tcp", addr)
+	conn, err := srv.newTCPListener(addr)
 	if err != nil {
 		return err
 	}
 
+	srv.TLSConfig = config
+
 	tlsListener := tls.NewListener(conn, config)
 	return srv.Serve(tlsListener)
 }
@@ -212,6 +251,7 @@ func (srv *Server) ListenAndServeTLSConfig(config *tls.Config) error {
 // If timeout is 0, the server never times out. It waits for all active requests to finish.
 func Serve(server *http.Server, l net.Listener, timeout time.Duration) error {
 	srv := &Server{Timeout: timeout, Server: server, Logger: DefaultLogger()}
+
 	return srv.Serve(l)
 }
 
@@ -219,20 +259,33 @@ func Serve(server *http.Server, l net.Listener, timeout time.Duration) error {
 func (srv *Server) Serve(listener net.Listener) error {
 
 	if srv.ListenLimit != 0 {
-		listener = netutil.LimitListener(listener, srv.ListenLimit)
+		listener = LimitListener(listener, srv.ListenLimit)
 	}
 
+	// Make our stopchan
+	srv.StopChan()
+
 	// Track connection state
 	add := make(chan net.Conn)
+	idle := make(chan net.Conn)
+	active := make(chan net.Conn)
 	remove := make(chan net.Conn)
 
 	srv.Server.ConnState = func(conn net.Conn, state http.ConnState) {
 		switch state {
 		case http.StateNew:
 			add <- conn
+		case http.StateActive:
+			active <- conn
+		case http.StateIdle:
+			idle <- conn
 		case http.StateClosed, http.StateHijacked:
 			remove <- conn
 		}
+
+		srv.stopLock.Lock()
+		defer srv.stopLock.Unlock()
+
 		if srv.ConnState != nil {
 			srv.ConnState(conn, state)
 		}
@@ -241,7 +294,7 @@ func (srv *Server) Serve(listener net.Listener) error {
 	// Manage open connections
 	shutdown := make(chan chan struct{})
 	kill := make(chan struct{})
-	go srv.manageConnections(add, remove, shutdown, kill)
+	go srv.manageConnections(add, idle, active, remove, shutdown, kill)
 
 	interrupt := srv.interruptChan()
 	// Set up the interrupt handler
@@ -306,28 +359,47 @@ func DefaultLogger() *log.Logger {
 	return log.New(os.Stderr, "[graceful] ", 0)
 }
 
-func (srv *Server) manageConnections(add, remove chan net.Conn, shutdown chan chan struct{}, kill chan struct{}) {
+func (srv *Server) manageConnections(add, idle, active, remove chan net.Conn, shutdown chan chan struct{}, kill chan struct{}) {
 	var done chan struct{}
 	srv.connections = map[net.Conn]struct{}{}
+	srv.idleConnections = map[net.Conn]struct{}{}
 	for {
 		select {
 		case conn := <-add:
 			srv.connections[conn] = struct{}{}
+		case conn := <-idle:
+			srv.idleConnections[conn] = struct{}{}
+		case conn := <-active:
+			delete(srv.idleConnections, conn)
 		case conn := <-remove:
 			delete(srv.connections, conn)
+			delete(srv.idleConnections, conn)
 			if done != nil && len(srv.connections) == 0 {
 				done <- struct{}{}
 				return
 			}
 		case done = <-shutdown:
-			if len(srv.connections) == 0 {
+			if len(srv.connections) == 0 && len(srv.idleConnections) == 0 {
 				done <- struct{}{}
 				return
 			}
+			// a shutdown request has been received. if we have open idle
+			// connections, we must close all of them now. this prevents idle
+			// connections from holding the server open while waiting for them to
+			// hit their idle timeout.
+			for k := range srv.idleConnections {
+				if err := k.Close(); err != nil {
+					srv.logf("[ERROR] %s", err)
+				}
+			}
 		case <-kill:
+			srv.stopLock.Lock()
+			defer srv.stopLock.Unlock()
+
+			srv.Server.ConnState = nil
 			for k := range srv.connections {
 				if err := k.Close(); err != nil {
-					srv.log("[ERROR] %s", err)
+					srv.logf("[ERROR] %s", err)
 				}
 			}
 			return
@@ -349,19 +421,22 @@ func (srv *Server) interruptChan() chan os.Signal {
 func (srv *Server) handleInterrupt(interrupt chan os.Signal, quitting chan struct{}, listener net.Listener) {
 	for _ = range interrupt {
 		if srv.Interrupted {
-			srv.log("already shutting down")
+			srv.logf("already shutting down")
 			continue
 		}
-		srv.log("shutdown initiated")
+		srv.logf("shutdown initiated")
 		srv.Interrupted = true
 		if srv.BeforeShutdown != nil {
-			srv.BeforeShutdown()
+			if !srv.BeforeShutdown() {
+				srv.Interrupted = false
+				continue
+			}
 		}
 
 		close(quitting)
 		srv.SetKeepAlivesEnabled(false)
 		if err := listener.Close(); err != nil {
-			srv.log("[ERROR] %s", err)
+			srv.logf("[ERROR] %s", err)
 		}
 
 		if srv.ShutdownInitiated != nil {
@@ -370,9 +445,11 @@ func (srv *Server) handleInterrupt(interrupt chan os.Signal, quitting chan struc
 	}
 }
 
-func (srv *Server) log(fmt string, v ...interface{}) {
-	if srv.Logger != nil {
-		srv.Logger.Printf(fmt, v...)
+func (srv *Server) logf(format string, args ...interface{}) {
+	if srv.LogFunc != nil {
+		srv.LogFunc(format, args...)
+	} else if srv.Logger != nil {
+		srv.Logger.Printf(format, args...)
 	}
 }
 
@@ -397,3 +474,14 @@ func (srv *Server) shutdown(shutdown chan chan struct{}, kill chan struct{}) {
 	}
 	srv.chanLock.Unlock()
 }
+
+func (srv *Server) newTCPListener(addr string) (net.Listener, error) {
+	conn, err := net.Listen("tcp", addr)
+	if err != nil {
+		return conn, err
+	}
+	if srv.TCPKeepAlive != 0 {
+		conn = keepAliveListener{conn, srv.TCPKeepAlive}
+	}
+	return conn, nil
+}
